/*
 * ccl-rusage.c -- 
 * 
 * This file is a part of the C Common Library (CCL) project. 
 * 
 * See file AUTHORS and COPYING for copyright details.
 *
 */
#include <math.h>
#include "ccl-rusage.h"
#if CCL_HAVE_GETRUSAGE
# include <sys/time.h>
# include <sys/resource.h>


int
ccl_getrusage (long *p_usertime, long *p_maxrss)
{
  int res;
  struct rusage r;
  
  if (getrusage (RUSAGE_SELF, &r) == 0)
    {
      if (p_usertime != NULL)
	*p_usertime = (long)
	  ceil (1000.0 * r.ru_utime.tv_sec + r.ru_utime.tv_usec / 1000.0);
      if (p_maxrss != NULL)
	*p_maxrss = r.ru_maxrss;
      res = 1;
    }
  else
    res = 0;
  
  return res;
}
#else
int 
ccl_getrusage (long *p_usertime, long *p_maxrss)
{
  if (p_usertime)
    *p_usertime = (long)
      ceil (1000.0 * (double) clock () / (double) CLOCKS_PER_SEC);
  if (p_maxrss)
    *p_maxrss = 0;
  
  return 0;  
}
#endif



