/*
 * group.h -- Groups
 * 
 * This file is a part of the Talence Linear Algebra Library. 
 * 
 * Copyright (C) 2010 CNRS UMR 5800 & Université Bordeaux I (see AUTHORS file).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

/*!
 * \file
 * \brief
 * 
 */
#ifndef __GROUP_H__
# define __GROUP_H__

# include <talalib/matrix.h>
# include <talalib/vsp.h>

BEGIN_C_DECLS

/* 
 * A group is given by an hermite square matrix with size*(size+1) coefficients
 * size+1 allow the add a new vector to the group easilly
 * dim is the dimension of the vector space generated by the group.
 * Thus only the first dim columns are non-zero. 
 */
typedef struct tla_group_st {
  int dim;
  tla_matrix_t Hermite;
} tla_group_t[1];

			/* --------------- */

extern void
tla_group_init(tla_group_t G, int size);
extern void
tla_group_init_set(tla_group_t res, const tla_group_t G);
extern void
tla_group_set(tla_group_t res, const tla_group_t G);
extern void
tla_group_clear(tla_group_t G);
extern void
tla_group_Z_cap_vsp(tla_group_t G, const tla_vsp_t vsp);
extern void
tla_group_z_mul (tla_group_t R, const tla_group_t G, const mpz_t scale);
extern void
tla_group_z_mul_ui (tla_group_t R, const tla_group_t G, unsigned int scale);

extern void
tla_group_add_generator(tla_group_t res, const tla_group_t G, 
			const tla_vec_t x);
extern int
tla_group_has(tla_group_t G, const tla_vec_t x, tla_vec_t lambda);
extern unsigned int
tla_group_hash (const tla_group_t G);
extern int
tla_group_are_equal (const tla_group_t G, const tla_group_t GO);
# define tla_group_get_size(G) (tla_matrix_get_nb_lines((G)->Hermite))
# define tla_group_get_dim(G) ((G)->dim)

extern int
tla_group_alpha_min(const tla_group_t G, const tla_vec_t alpha, const mpz_t c, 
		    tla_vec_t x);

END_C_DECLS

#endif /* ! __GROUP_H__ */
